from .base import MyApiTestCase
import json
from . import radiusmock
from privacyidea.lib.config import set_privacyidea_config
from privacyidea.lib.radiusserver import delete_radius
DICT_FILE = "tests/testdata/dictionary"


class RADIUSServerTestCase(MyApiTestCase):
    """
    test the api.radiusserver endpoints
    """

    def test_01_create_server(self):
        # create and list server

        # Unauthorized
        with self.app.test_request_context('/radiusserver/server1',
                                           method='POST',
                                           data={"secret": "testing123",
                                                 "port": "1812",
                                                 "server": "1.2.3.4",
                                                 "description": "myServer"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/radiusserver/server1',
                                           method='POST',
                                           data={"secret": "testing123",
                                                 "port": "1812",
                                                 "server": "1.2.3.4",
                                                 "description": "myServer"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            self.assertEqual(data.get("result").get("value"), True)

        # list servers
        with self.app.test_request_context('/radiusserver/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            server_list = data.get("result").get("value")
            self.assertEqual(len(server_list), 1)
            server1 = server_list.get("server1")
            self.assertEqual(server1.get("server"), "1.2.3.4")
            self.assertEqual(server1.get("description"), "myServer")
            self.assertTrue("secret" not in server1)

        # delete server
        with self.app.test_request_context('/radiusserver/server1',
                                           method='DELETE',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        # list servers, No server left
        with self.app.test_request_context('/radiusserver/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            server_list = data.get("result").get("value")
            self.assertEqual(len(server_list), 0)

    @radiusmock.activate
    def test_02_send_test_email(self):
        set_privacyidea_config("radius.dictfile", DICT_FILE)
        radiusmock.setdata(response=radiusmock.AccessAccept)

        with self.app.test_request_context('/radiusserver/test_request',
                                           method='POST',
                                           data={"identifier": "someServer",
                                                 "secret": "secret",
                                                 "port": "1812",
                                                 "server": "1.2.3.4",
                                                 "dictionary": DICT_FILE,
                                                 "username": "testuser",
                                                 "password": "testpassword"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            self.assertEqual(data.get("result").get("value"), True)

    def test_03_radiusserver_user(self):
        # The user must be able to call GET /radiusserver/
        # But not POST and not DELETE
        # delete server
        self.setUp_user_realms()
        self.authenticate_selfservice_user()

        # User is not allowed to delete a radiusserver
        with self.app.test_request_context('/radiusserver/server1',
                                           method='DELETE',
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        # User is not allowed to create a radius server
        with self.app.test_request_context('/radiusserver/server1',
                                           method='POST',
                                           data={"secret": "testing123",
                                                 "port": "1812",
                                                 "server": "1.2.3.4",
                                                 "description": "myServer"},
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 401, res)

        with self.app.test_request_context('/radiusserver/server1',
                                           method='POST',
                                           data={"secret": "testing123",
                                                 "port": "1812",
                                                 "server": "1.2.3.4",
                                                 "description": "myServer"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            data = res.json
            self.assertEqual(data.get("result").get("value"), True)

        # Users are not allowed to list the radius servers
        with self.app.test_request_context('/radiusserver/',
                                           method='GET',
                                           headers={'Authorization': self.at_user}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401)
            result = res.json.get("result")
            self.assertIn("do not have the necessary role", result["error"]["message"])

        delete_radius("server1")



#  2018-06-26 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             provide first interface to handle monitoring statistics
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This module is used to write, read and handle data in the
database table "monitoringstats". This can be arbitrary data for time series of
  
   timestamp, key, value

This module is tested in tests/test_lib_monitoringstats.py
"""
import logging
from dateutil.tz import tzlocal
from privacyidea.lib.log import log_with
from privacyidea.lib.utils import get_module_class
from privacyidea.lib.framework import get_app_config, get_request_local_store
import datetime

log = logging.getLogger(__name__)


@log_with(log, log_entry=False)
def _get_monitoring():
    """
    This wrapper function creates a new monitoring object based on the config
    from the config file. The config file entry could look like this:

        PI_MONITORING_MODULE = privacyidea.lib.monitoringmodule.sqlstats

    Each monitoring module can have its own config values.

    :return: Monitoring Object
    """
    store = get_request_local_store()
    # check if the monitoring object is not yet in this request_store
    if 'monitoring_object' not in store:
        config = get_app_config()
        monitoring_module = config.get("PI_MONITORING_MODULE", "privacyidea.lib.monitoringmodules.sqlstats")
        monitoring = get_module_class(monitoring_module, "Monitoring")(config)
        store['monitoring_object'] = monitoring
    return store.get('monitoring_object')


def write_stats(stats_key, stats_value, timestamp=None, reset_values=False):
    """
    Write a new statistics value to the database

    :param stats_key: The key, that identifies the measurement point
    :type stats_key: basestring
    :param stats_value: The value to be measured
    :type stats_value: int
    :param timestamp: The time, when the value was measured
    :type timestamp: timezone-aware datetime object
    :param reset_values: Whether old entries should be deleted
    :return: id of the database entry
    """
    timestamp = timestamp or datetime.datetime.now(tzlocal())
    monitoring_obj = _get_monitoring()
    monitoring_obj.add_value(stats_key, stats_value, timestamp, reset_values)


def delete_stats(stats_key, start_timestamp=None, end_timestamp=None):
    """
    Delete statistics from a given key.
    Either delete all occurrences or only in a given time span.

    :param stats_key: The name of the key to delete
    :param start_timestamp: The start timestamp.
    :type start_timestamp: timezone-aware datetime object
    :param end_timestamp: The end timestamp.
    :type end_timestamp: timezone-aware datetime object
    :return: The number of deleted entries
    """
    monitoring_obj = _get_monitoring()
    r = monitoring_obj.delete(stats_key, start_timestamp, end_timestamp)
    return r


def get_stats_keys():
    """
    Return a list of all available statistics keys

    :return: list of keys
    """
    monitoring_obj = _get_monitoring()
    return monitoring_obj.get_keys()


def get_values(stats_key, start_timestamp=None, end_timestamp=None):
    """
    Return a list of sets of (timestamp, value), ordered by timestamps in ascending order

    :param stats_key: The stats key to query
    :param start_timestamp: the start of the timespan, inclusive
    :type start_timestamp: timezone-aware datetime object
    :param end_timestamp: the end of the timespan, inclusive
    :type end_timestamp: timezone-aware datetime object
    :return: list of tuples, with timestamps being timezone-aware UTC datetime objects
    """
    monitoring_obj = _get_monitoring()
    return monitoring_obj.get_values(stats_key, start_timestamp, end_timestamp)


def get_last_value(stats_key):
    """
    Return the last value of the given key

    :param stats_key:
    :return: The value as a scalar or None
    """
    monitoring_obj = _get_monitoring()
    return monitoring_obj.get_last_value(stats_key)




"""
2019-08-15 More sophisticated radiusmock
        Cornelius Kölbel <cornelius.koelbel@netknights.it>
2017-10-30 Add mocking the timeout
        Cornelius Kölbel <cornelius.koelbel@netknights.it>
2015-01-29 Change responses.py to be able to run with RADIUS
        Cornelius Kölbel <cornelius@privacyidea.org>

Original responses.py is:
Copyright 2013 Dropbox, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import namedtuple

from collections.abc import Sequence, Sized

from pyrad import packet
from pyrad.client import Timeout
from pyrad.packet import AccessReject, AccessAccept, AccessChallenge

from .smtpmock import get_wrapped

Call = namedtuple('Call', ['request', 'response'])

_wrapper_template = """\
def wrapper%(signature)s:
    with radiusmock:
        return func%(funcargs)s
"""


class CallList(Sequence, Sized):
    def __init__(self):
        self._calls = []

    def __iter__(self):
        return iter(self._calls)

    def __len__(self):
        return len(self._calls)

    def __getitem__(self, idx):
        return self._calls[idx]

    def setdata(self, request, response):
        self._calls.append(Call(request, response))

    def reset(self):
        self._calls = []


class RadiusMock(object):

    def __init__(self):
        self._calls = CallList()
        self.reset()

    def reset(self):
        self._request_data = {}
        self._calls.reset()

    def setdata(self, server=None, rpacket=None, response=AccessReject, response_data=None, timeout=False):
        self._request_data = {
            'server': server,
            'packet': rpacket,
            'response': response,
            'response_data': response_data or {},
            'timeout': timeout
        }

    @property
    def calls(self):
        return self._calls

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop()
        self.reset()

    def activate(self, func):
        evaldict = {'radiusmock': self, 'func': func}
        return get_wrapped(func, _wrapper_template, evaldict)

    def _on_request(self, client_instance, pkt):
        # mangle request packet
        request = pkt.RequestPacket()
        if pkt.code == packet.AccessRequest:
            # This is a request
            pass
        # create reply packet
        """
                1       Access-Request
                2       Access-Accept
                3       Access-Reject
                4       Accounting-Request
                5       Accounting-Response
               11       Access-Challenge
               12       Status-Server (experimental)
               13       Status-Client (experimental)
              255       Reserved
        """
        #reply = pkt.CreateReply(packet=rawreply)
        if self._request_data.get("timeout"):
            raise Timeout()
        reply = pkt.CreateReply(**self._request_data.get("response_data"))
        reply.code = self._request_data.get("response", AccessReject)
        return reply

    def start(self):
        import mock

        def unbound_on_send(Client, pkt, *a, **kwargs):
            return self._on_request(Client, pkt,  *a, **kwargs)
        self._patcher = mock.patch('pyrad.client.Client.SendPacket',
                                   unbound_on_send)
        self._patcher.start()

    def stop(self):
        self._patcher.stop()


# expose default mock namespace
mock = _default_mock = RadiusMock()
__all__ = []
for __attr in (a for a in dir(_default_mock) if not a.startswith('_')):
    __all__.append(__attr)
    globals()[__attr] = getattr(_default_mock, __attr)



import json
from .base import MyApiTestCase
from privacyidea.lib.clientapplication import save_clientapplication


class APIClienttypeTestCase(MyApiTestCase):

    def test_00_get_client(self):
        save_clientapplication("1.2.3.4", "PAM")
        save_clientapplication("1.2.3.4", "RADIUS")
        with self.app.test_request_context('/client/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            json_response = res.json
            self.assertTrue(json_response.get("result").get("status"), res)
            self.assertTrue("PAM" in json_response.get("result").get("value"))
            self.assertTrue("RADIUS" in json_response.get("result").get(
                "value"))

        # test the decorator
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           data={"user": "hans",
                                                 "pass": "dampf"},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            # Some user... does not exist
            self.assertTrue(res.status_code == 400, res)

        with self.app.test_request_context('/client/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            json_response = res.json
            self.assertTrue(json_response.get("result").get("status"), res)
            self.assertTrue("PAM" in json_response.get("result").get("value"))
            self.assertTrue("RADIUS" in json_response.get("result").get(
                "value"))
            self.assertTrue("unknown" in json_response.get("result").get(
                "value"))



# http://www.privacyidea.org
# (c) cornelius kölbel, privacyidea.org
#
# 2017-07-11 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Allow to return GPG keys
# 2015-12-18 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Remove the complete before and after logic
# 2015-10-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add a call for testing token config
# 2015-09-25 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add HSM interface
# 2015-06-26 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#            Add system config documentation API
# 2014-12-08 Cornelius Kölbel, <cornelius@privacyidea.org>
#            Complete rewrite during flask migration
#            Try to provide REST API
#
# privacyIDEA is a fork of LinOTP. Some code is adapted from
# the system-controller from LinOTP, which is
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  AGPLv3
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """
This is the REST API for system calls to create and read system configuration.

The code of this module is tested in tests/test_api_system.py
"""

from flask import (Blueprint,
                   request)
from .lib.utils import (getParam,
                        getLowerParams,
                        optional,
                        required,
                        send_result, send_file)
from ..lib.log import log_with
from ..lib.radiusserver import get_radiusservers
from ..lib.caconnector import get_caconnector_list
from ..lib.config import (get_token_class,
                          set_privacyidea_config,
                          delete_privacyidea_config,
                          get_from_config,
                          get_privacyidea_nodes)
from ..api.lib.prepolicy import prepolicy, check_base_action
from ..lib.error import ParameterError

from .auth import admin_required
from flask import (g, current_app, render_template)
import logging
import datetime
import re
import socket
from privacyidea.lib.resolver import get_resolver_list
from privacyidea.lib.realm import get_realms
from privacyidea.lib.policy import PolicyClass, ACTION
from privacyidea.lib.auth import get_db_admins
from privacyidea.lib.crypto import geturandom, set_hsm_password, get_hsm
from privacyidea.lib.importotp import GPGImport
from privacyidea.lib.utils import hexlify_and_unicode, b64encode_and_unicode


log = logging.getLogger(__name__)


system_blueprint = Blueprint('system_blueprint', __name__)


@system_blueprint.route('/documentation', methods=['GET'])
@admin_required
@prepolicy(check_base_action, request, ACTION.CONFIGDOCUMENTATION)
def get_config_documentation():
    """
    returns an restructured text document, that describes the complete
    configuration.
    """
    P = PolicyClass()

    config = get_from_config()
    resolvers = get_resolver_list()
    realms = get_realms()
    policies = P.list_policies()
    admins = get_db_admins()
    context = {"system": socket.getfqdn(socket.gethostname()),
               "date": datetime.datetime.now().strftime("%Y-%m-%d %H:%M"),
               "systemconfig": config,
               "appconfig": current_app.config,
               "resolverconfig": resolvers,
               "realmconfig": realms,
               "policyconfig": policies,
               "admins": admins}

    g.audit_object.log({"success": True})
    # Three or more line breaks will be changed to two.
    return send_file(re.sub("\n{3,}", "\n\n", render_template("documentation.rst",
                                                              context=context)),
                     'documentation.rst', content_type='text/plain')


@system_blueprint.route('/gpgkeys', methods=['GET'])
def get_gpg_keys():
    """
    Returns the GPG keys in the config directory specified by PI_GNUPG_HOME.

    :return: A json list of the public GPG keys
    """
    GPG = GPGImport(current_app.config)
    keys = GPG.get_publickeys()
    g.audit_object.log({"success": True})
    return send_result(keys)


@system_blueprint.route('/<key>', methods=['GET'])
@system_blueprint.route('/', methods=['GET'])
def get_config(key=None):
    """
    This endpoint either returns all config entries or only the value of the
    one config key.

    This endpoint can be called by the administrator but also by the normal
    user, so that the normal user gets necessary information about the system
    config

    :param key: (optional) The key to return
    :>json bool status: Status of the request
    :>json value: JSON object with a key-value pair of the config entries or
    :>json value: The value of the specified config entry
    :reqheader PI-Authorization: The authorization token

    **Example request 1**:

    .. sourcecode:: http

       GET /system/ HTTP/1.1
       Host: example.com
       Content-Type: application/json

    **Example response 1**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "AutoResync": "False",
              "splitAtSign": "True",
              "PrependPin": "True",
              "DefaultCountWindow": "10"
            }
          },
          "version": "privacyIDEA unknown"
        }

    **Example request 2**:
    Querying a specific system-configuration value

    .. sourcecode:: http

       GET /system/totp.hashlib HTTP/1.1
       Host: example.com
       Content-Type: application/json

    **Example response 2**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": "sha1"
          },
          "version": "privacyIDEA unknown"
        }
    """
    if not key:
        result = get_from_config(role=g.logged_in_user.get("role"))
    else:
        result = get_from_config(key, role=g.logged_in_user.get("role"))

    g.audit_object.log({"success": True,
                        "info": key})
    return send_result(result)


@system_blueprint.route('/setConfig', methods=['POST'])
@admin_required
@prepolicy(check_base_action, request, ACTION.SYSTEMWRITE)
def set_config():
    """
    set a configuration key or a set of configuration entries

    parameter are generic ``keyname=value`` pairs.

    **remark** In case of key-value pairs the type information could be
        provided by an additional parameter with same keyname with the
        postfix ".type". Value could then be 'password' to trigger the
        storing of the value in an encrypted form

    :<json key-value-pairs: a list of ``keyname=value`` pairs
    :<json <keyname>.type: type of the value: int or string/text or password.
        password will trigger to store the encrypted value
    :<json <keyname>.desc: additional information for this config entry
    :>json bool status: Status of the request
    :>json value: JSON object with a list of key-value pairs of the requested
        config entry changes with the value of ``update`` or ``insert``
    :reqheader PI-Authorization: The authorization token

    **Example request**:

    .. sourcecode:: http

        POST /system/setConfig HTTP/1.1
        Host: example.com
        Content-Type: application/json

        "splitAtSign": true
        "totp.hashlib": "sha1"
        "totp.hashlib.desc": "The hash algorithm used for TOTP tokens"

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": {
              "splitAtSign": "update",
              "totp.hashlib": "update"
            }
          },
          "version": "privacyIDEA unknown"
        }
    """
    param = request.all_data
    result = {}
    for key in param:
        if key.split(".")[-1] not in ["type", "desc"]:
            # Only store base values, not type or desc
            value = getParam(param, key, optional)
            typ = getParam(param, key + ".type", optional)
            desc = getParam(param, key + ".desc", optional)
            res = set_privacyidea_config(key, value, typ, desc)
            result[key] = res
            g.audit_object.add_to_log({"info": "{0!s}={1!s}, ".format(key, value)})
    g.audit_object.log({"success": True})
    return send_result(result)


@system_blueprint.route('/setDefault', methods=['POST'])
@admin_required
@prepolicy(check_base_action, request, ACTION.SYSTEMWRITE)
def set_default():
    """
    define default settings for tokens. These default settings
    are used when new tokens are generated. The default settings will
    not affect already enrolled tokens.

    :jsonparam DefaultMaxFailCount: Default value for the maximum allowed
        authentication failures
    :jsonparam DefaultSyncWindow: Default value for the synchronization window
    :jsonparam DefaultCountWindow: Default value for the counter window
    :jsonparam DefaultOtpLen: Default value for the OTP value length --
        usually 6 or 8
    :jsonparam DefaultResetFailCount: Default value, if the FailCounter should
        be reset on successful authentication [True|False]

    :return: a json result with a boolean "result": true

    """
    keys = ["DefaultMaxFailCount",
            "DefaultSyncWindow",
            "DefaultCountWindow",
            "DefaultOtpLen",
            "DefaultResetFailCount"]

    description = "parameters are: {0!s}".format(", ".join(keys))
    param = getLowerParams(request.all_data)
    result = {}
    for k in keys:
        if k.lower() in param:
            value = getParam(param, k.lower(), required)
            res = set_privacyidea_config(k, value)
            result[k] = res
            g.audit_object.log({"success": True})
            g.audit_object.add_to_log({"info": "{0!s}={1!s}, ".format(k, value)})

    if not result:
        log.warning("Failed saving config. Could not find any "
                    "known parameter. %s"
                    % description)
        raise ParameterError("Usage: {0!s}".format(description), id=77)

    return send_result(result)


@system_blueprint.route('/<key>', methods=['DELETE'])
@admin_required
@prepolicy(check_base_action, request, ACTION.SYSTEMDELETE)
@log_with(log)
def delete_config(key=None):
    """
    delete a configuration key

    :param string key: configuration key name
    :returns: a json result with the deleted value

    """
    if not key:
        raise ParameterError("You need to provide the config key to delete.")
    res = delete_privacyidea_config(key)
    g.audit_object.log({'success': res,
                        'info': key})
    return send_result(res)


@system_blueprint.route('/hsm', methods=['POST'])
@admin_required
@prepolicy(check_base_action, request, ACTION.SETHSM)
@log_with(log)
def set_security_module():
    """
    Set the password for the security module
    """
    password = getParam(request.all_data, "password", required)
    is_ready = set_hsm_password(password)
    res = {"is_ready": is_ready}
    g.audit_object.log({'success': res})
    return send_result(res)


@system_blueprint.route('/hsm', methods=['GET'])
@admin_required
@log_with(log)
def get_security_module():
    """
    Get the status of the security module.
    """
    hsm = get_hsm(require_ready=False)
    is_ready = hsm.is_ready
    res = {"is_ready": is_ready}
    g.audit_object.log({'success': res})
    return send_result(res)


@system_blueprint.route('/random', methods=['GET'])
@admin_required
@prepolicy(check_base_action, request, action=ACTION.GETRANDOM)
@log_with(log)
def rand():
    """
    This endpoint can be used to retrieve random keys from privacyIDEA.
    In certain cases the client might need random data to initialize tokens
    on the client side. E.g. the command line client when initializing the
    yubikey or the WebUI when creating Client API keys for the yubikey.

    In this case, privacyIDEA can create the random data/keys.

    :queryparam len: The length of a symmetric key (byte)
    :queryparam encode: The type of encoding. Can be "hex" or "b64".

    :return: key material
    """
    length = int(getParam(request.all_data, "len") or 20)
    encode = getParam(request.all_data, "encode")

    r = geturandom(length=length)
    if encode == "b64":
        res = b64encode_and_unicode(r)
    else:
        res = hexlify_and_unicode(r)

    g.audit_object.log({'success': res})
    return send_result(res)


@system_blueprint.route('/test/<tokentype>', methods=['POST'])
@admin_required
@prepolicy(check_base_action, request, action=ACTION.SYSTEMWRITE)
@log_with(log)
def test(tokentype=None):
    """
    The call /system/test/email tests the configuration of the email token.
    """
    tokenc = get_token_class(tokentype)
    res, description = tokenc.test_config(request.all_data)
    g.audit_object.log({"success": 1,
                        "token_type": tokentype})
    return send_result(res, details={"message": description})


@system_blueprint.route('/names/radius', methods=['GET'])
@prepolicy(check_base_action, request, action="enrollRADIUS")
def list_radius_servers():
    """
    Return the list of identifiers of all defined RADIUS servers.
    This endpoint requires the enrollRADIUS right.
    """
    server_list = get_radiusservers()
    res = [server.config.identifier for server in server_list]
    g.audit_object.log({'success': True})
    return send_result(res)


@system_blueprint.route('/names/caconnector', methods=['GET'])
@prepolicy(check_base_action, request, action="enrollCERTIFICATE")
def list_ca_connectors():
    """
    Return a list of defined CA connectors. Each item of the list is
    a dictionary with the CA connector information, including the
    name and defined templates, but excluding the CA connector data.
    This endpoint requires the enrollCERTIFICATE right.
    """
    ca_list = get_caconnector_list(return_config=False)
    g.audit_object.log({"success": True})
    return send_result(ca_list)


@system_blueprint.route("/nodes", methods=['GET'])
@admin_required
def list_nodes():
    """
    Return a list of nodes, that are known to the system.

    :>json list nodes: A list of JSON objects with the node name and uuid
    :reqheader PI-Authorization: The authorization token

    **Example response**:

    .. sourcecode:: http

        HTTP/1.1 200 OK
        Content-Type: application/json

        {
          "id": 1,
          "jsonrpc": "2.0",
          "result": {
            "status": true,
            "value": [
                {
                    "name": "node1",
                    "uuid": "12345678-1234-1234-1234-1234567890ab"
                },
                {
                    "name": "node2",
                    "uuid": "12345678-4321-1234-1234-1234567890ac"
                }
            ]
          },
          "version": "privacyIDEA unknown"
        }
    .. versionadded:: 3.10 Return node information with names and UUIDs
    """
    nodes = get_privacyidea_nodes()
    g.audit_object.log({"success": True})
    return send_result(nodes)



#  2016-04-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Avoid consecutive if-statements
#  2015-12-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Change eval to importlib
#  2015-04-23 Cornelius Kölbel <cornelius.koelbel@netknigts.it>
#             Add CA Connector functions
#
#  privacyIDEA is a fork of LinOTP
#  Nov 11, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """The config module takes care about storing server configuration in
the Config database table.

It provides functions to retrieve (get) and and set configuration.

The code is tested in tests/test_lib_config
"""

import copy
import sys
import logging
import inspect
import threading
import traceback

from .log import log_with
from ..models import (Config, db, Resolver, Realm, PRIVACYIDEA_TIMESTAMP,
                      save_config_timestamp, Policy, EventHandler, CAConnector,
                      NodeName)
from privacyidea.lib.framework import get_request_local_store, get_app_config_value, get_app_local_store
from privacyidea.lib.utils import to_list
from privacyidea.lib.utils.export import (register_import, register_export)
from .crypto import encryptPassword
from .crypto import decryptPassword
from .resolvers.UserIdResolver import UserIdResolver
from .machines.base import BaseMachineResolver
from .caconnectors.baseca import BaseCAConnector
# We need these imports to return the list of CA connector types. Bummer: New import for each new Class anyway.
from .caconnectors import localca, msca
from .utils import reload_db, is_true
import importlib
import datetime

log = logging.getLogger(__name__)

ENCODING = 'utf-8'

# this is a pointer to the module object instance itself.
this = sys.modules[__name__]

this.config = {}


class SharedConfigClass(object):
    """
    A shared config class object is shared between threads and is supposed
    to store the current configuration with resolvers, realms, policies
    and event handler definitions along with the timestamp of the configuration.

    The method ``_reload_from_db()`` compares this timestamp against the
    timestamp in the database (while taking the PI_CHECK_RELOAD_CONFIG
    setting into account). If the database timestamp is newer, the current
    configuration is updated.

    However, app code must not access the config stored in the shared object!
    Instead, it must use ``reload_and_clone()`` to retrieve
    a ``LocalConfigClass`` object which holds a local configuration snapshot.
    """

    def __init__(self):
        self._config_lock = threading.Lock()
        self.config = {}
        self.resolver = {}
        self.realm = {}
        self.default_realm = None
        self.policies = []
        self.events = []
        self.timestamp = None
        self.caconnectors = []

    def _reload_from_db(self):
        """
        Read the timestamp from the database. If the timestamp is newer than
        the internal timestamp, then read the complete data
        :return:
        """
        check_reload_config = get_app_config_value("PI_CHECK_RELOAD_CONFIG", 0)
        if not self.timestamp or \
                self.timestamp + datetime.timedelta(seconds=check_reload_config) < datetime.datetime.now():
            db_ts = Config.query.filter_by(Key=PRIVACYIDEA_TIMESTAMP).first()
            if reload_db(self.timestamp, db_ts):
                log.debug("Reloading shared config from database")
                config = {}
                resolverconfig = {}
                realmconfig = {}
                default_realm = None
                policies = []
                events = []
                caconnectors = []
                # Load system configuration
                for sysconf in Config.query.all():
                    config[sysconf.Key] = {
                        "Value": sysconf.Value,
                        "Type": sysconf.Type,
                        "Description": sysconf.Description}
                # Load resolver configuration
                for resolver in Resolver.query.all():
                    resolverdef = {"type": resolver.rtype,
                                   "resolvername": resolver.name,
                                   "censor_keys": []}
                    data = {}
                    for rconf in resolver.config_list:
                        if rconf.Type == "password":
                            value = decryptPassword(rconf.Value)
                            resolverdef["censor_keys"].append(rconf.Key)
                        else:
                            value = rconf.Value
                        data[rconf.Key] = value
                    resolverdef["data"] = data
                    resolverconfig[resolver.name] = resolverdef
                # Load realm configuration
                for realm in Realm.query.all():
                    if realm.default:
                        default_realm = realm.name
                    realmdef = {"id": realm.id,
                                "option": realm.option,
                                "default": realm.default,
                                "resolver": []}
                    for x in realm.resolver_list:
                        realmdef["resolver"].append({"priority": x.priority,
                                                     "name": x.resolver.name,
                                                     "type": x.resolver.rtype,
                                                     "node": x.node_uuid})
                    realmconfig[realm.name] = realmdef
                # Load all policies
                for pol in Policy.query.all():
                    policies.append(pol.get())
                # Load all events
                for event in EventHandler.query.order_by(EventHandler.ordering):
                    events.append(event.get())
                # Load all CA connectors
                from privacyidea.lib.caconnector import get_caconnector_object
                for ca in CAConnector.query.all():
                    try:
                        ca_obj = get_caconnector_object(ca.name)
                        caconnectors.append({"connectorname": ca.name,
                                             "type": ca.catype,
                                             "data": ca_obj.config,
                                             "templates": ca_obj.get_templates()})
                    except Exception as exx:  # pragma: no cover
                        log.debug("{0!s}".format(traceback.format_exc()))
                        log.error(exx)

                # Finally, set the current timestamp
                timestamp = datetime.datetime.now()
                with self._config_lock:
                    self.config = config
                    self.resolver = resolverconfig
                    self.realm = realmconfig
                    self.default_realm = default_realm
                    self.policies = policies
                    self.events = events
                    self.timestamp = timestamp
                    self.caconnectors = caconnectors

    def _clone(self):
        """
        :return: a ``LocalConfigClass`` object containing the current configuration state
        """
        with self._config_lock:
            return LocalConfigClass(
                self.config,
                self.resolver,
                self.realm,
                self.default_realm,
                self.policies,
                self.events,
                self.caconnectors,
                self.timestamp
            )

    def reload_and_clone(self):
        """
        Check if the current configuration state is outdated (according to the
        PI_CHECK_RELOAD_CONFIG setting), reload it if needed and return a
        ``LocalConfigClass`` object containing the current configuration state
        """
        self._reload_from_db()
        return self._clone()


class LocalConfigClass(object):
    """
    The Config_Object will contain all database configuration of system
    config, resolvers, realms, policies and event handler definitions.

    It will be cloned from the shared config object at the beginning of the
    request and is supposed to stay alive and unchanged during the request.
    """

    def __init__(self, config, resolver, realm, default_realm, policies, events, caconnectors, timestamp):
        self.config = config
        self.resolver = resolver
        self.realm = realm
        self.default_realm = default_realm
        self.policies = policies
        self.events = events
        self.caconnectors = caconnectors
        self.timestamp = timestamp

    def get_config(self, key=None, default=None, role="admin",
                   return_bool=False):
        """
        :param key: A key to retrieve
        :type key: string
        :param default: The default value, if it does not exist in the database
        :param role: The role which wants to retrieve the system config. Can be
            "admin" or "public". If "public", only values with type="public"
            are returned.
        :type role: string
        :param return_bool: If a boolean value should be returned. Returns
            True if value is "True", "true", 1, "1", True...
        :return: If key is None, then a dictionary is returned. If a certain key
            is given a string/bool is returned.
        """
        default_true_keys = [SYSCONF.PREPENDPIN, SYSCONF.SPLITATSIGN,
                             SYSCONF.INCFAILCOUNTER, SYSCONF.RETURNSAML]

        r_config = {}

        # reduce the dictionary to only public keys!
        reduced_config = {}
        for ckey, cvalue in self.config.items():
            if role == "admin" or cvalue.get("Type") == "public":
                reduced_config[ckey] = self.config[ckey]
        if not reduced_config and role == "admin":
            reduced_config = self.config

        for ckey, cvalue in reduced_config.items():
            if cvalue.get("Type") == "password":
                # decrypt the password
                r_config[ckey] = decryptPassword(cvalue.get("Value"))
            else:
                r_config[ckey] = cvalue.get("Value")

        for t_key in default_true_keys:
            if t_key not in r_config:
                r_config[t_key] = "True"

        if key:
            # We only return a single key
            r_config = r_config.get(key, default)

        if return_bool:
            if isinstance(r_config, bool):
                pass
            if isinstance(r_config, int):
                r_config = r_config > 0
            if isinstance(r_config, str):
                r_config = is_true(r_config.lower())

        return r_config


class SYSCONF(object):
    __doc__ = """This is a list of system config attributes"""
    OVERRIDECLIENT = "OverrideAuthorizationClient"
    PREPENDPIN = "PrependPin"
    SPLITATSIGN = "splitAtSign"
    INCFAILCOUNTER = "IncFailCountOnFalsePin"
    RETURNSAML = "ReturnSamlAttributes"
    RETURNSAMLONFAIL = "ReturnSamlAttributesOnFail"
    RESET_FAILCOUNTER_ON_PIN_ONLY = "ResetFailcounterOnPIN"


# @cache.cached(key_prefix="allConfig")
def get_privacyidea_config():
    # timestamp = Config.query.filter_by(Key="privacyidea.timestamp").first()
    return get_from_config()


def get_shared_config_object():
    """
    :return: the application-wide ``SharedConfigClass`` object, which is created on demand.
    """
    store = get_app_local_store()
    if 'shared_config_object' not in store:
        # It might happen that two threads create SharedConfigClass() instances in parallel.
        # However, as setting dictionary values is atomic, one of the two objects will "win",
        # and the next request handled by the second thread will use the winning config object.
        log.debug("Creating new shared config object")
        store['shared_config_object'] = SharedConfigClass()
    return store['shared_config_object']


def invalidate_config_object():
    """
    Invalidate the request-local config object. This is useful whenever a request modifies
    the configuration in the database: In this case, the request-local config object still
    contains the old configuration and needs to be invalidated. The same request can then
    create a new request-local config object, which will contain the new configuration.

    In other words, this function does the following:
    If the request-local store contains a config object, remove it.
    If the request-local store contains no config object, do nothing.
    """
    store = get_request_local_store()
    if 'config_object' in store:
        log.debug("Invalidating request-local config object")
        del store['config_object']


def ensure_no_config_object():
    """
    Ensure that the request-local store contains no config object, and emit
    a warning if it does. This should only be the case if we are running tests.

    If the request-local store contains a config object, remove it and emit a warning.
    If the request-local store contains no config object, do nothing.
    """
    store = get_request_local_store()
    if 'config_object' in store:
        log.warning("Request-local store already contains config object, even though it should not")
        del store['config_object']


def get_config_object():
    """
    Return the request-local config object. If it does not exist yet, create it.
    :return: a ``LocalConfigClass`` object
    """
    store = get_request_local_store()
    if 'config_object' not in store:
        log.debug("Cloning request-local config from shared config object")
        shared_config = get_shared_config_object()
        store['config_object'] = shared_config.reload_and_clone()
    return store['config_object']


@log_with(log)
# @cache.cached(key_prefix="singleConfig")
def get_from_config(key=None, default=None, role="admin", return_bool=False):
    """
    :param key: A key to retrieve
    :type key: string
    :param default: The default value, if it does not exist in the database
    :param role: The role which wants to retrieve the system config. Can be
        "admin" or "public". If "public", only values with type="public"
        are returned.
    :type role: string
    :param return_bool: Whether a boolean value should be returned. Returns
        True if value is "True", "true", 1, "1", True...
    :return: If key is None, then a dictionary is returned. If a certain key
        is given a string/bool is returned.
    """
    config_object = get_config_object()
    return config_object.get_config(key=key, default=default, role=role, return_bool=return_bool)


# @cache.cached(key_prefix="resolver")
def get_resolver_types():
    """
    Return a simple list of the type names of the resolvers.
    :return: array of resolvertypes like 'passwdresolver'
    :rtype: array
    """
    if "pi_resolver_types" not in this.config:
        (r_classes, r_types) = get_resolver_class_dict()
        this.config["pi_resolver_classes"] = r_classes
        this.config["pi_resolver_types"] = r_types

    return list(this.config["pi_resolver_types"].values())


def get_caconnector_types():
    """
    Returns a list of valid CA connector types
    :return:
    """
    caconnector_types = []
    for cacon in BaseCAConnector.__subclasses__():
        caconnector_types.append(cacon.connector_type)
    return caconnector_types


# @cache.cached(key_prefix="classes")
def get_resolver_classes():
    """
    Returns a list of the available resolver classes like:
    [<class 'privacyidea.lib.resolvers.PasswdIdResolver.IdResolver'>,
    <class 'privacyidea.lib.resolvers.UserIdResolver.UserIdResolver'>]

    :return: array of resolver classes
    :rtype: array
    """
    resolver_classes = {}
    if "pi_resolver_classes" not in this.config:
        (r_classes, r_types) = get_resolver_class_dict()
        this.config["pi_resolver_types"] = r_types
        this.config["pi_resolver_classes"] = r_classes

    return list(this.config["pi_resolver_classes"].values())


# @cache.cached(key_prefix="classes")
def get_token_class_dict():
    """
    get a dictionary of the token classes and a dictionary of the
    token types:

    ({'privacyidea.lib.tokens.hotptoken.HotpTokenClass':
      <class 'privacyidea.lib.tokens.hotptoken.HotpTokenClass'>,
      'privacyidea.lib.tokens.totptoken.TotpTokenClass':
      <class 'privacyidea.lib.tokens.totptoken.TotpTokenClass'>},

      {'privacyidea.lib.tokens.hotptoken.HotpTokenClass':
      'hotp',
      'privacyidea.lib.tokens.totptoken.TotpTokenClass':
      'totp'})

    :return: tuple of two dicts
    """
    from .tokenclass import TokenClass

    tokenclass_dict = {}
    tokentype_dict = {}
    modules = get_token_module_list()
    for module in modules:
        for name in dir(module):
            obj = getattr(module, name)
            # We must not process imported classes!
            if (inspect.isclass(obj) and issubclass(obj, TokenClass) and
                    obj.__module__ == module.__name__):
                try:
                    class_name = "{0!s}.{1!s}".format(module.__name__, obj.__name__)
                    tokenclass_dict[class_name] = obj
                    if hasattr(obj, 'get_class_type'):
                        tokentype_dict[class_name] = obj.get_class_type()
                except Exception as e:  # pragma: no cover
                    log.error("error constructing token_class_dict: {0!r}".format(e))

    return tokenclass_dict, tokentype_dict


# @cache.cached(key_prefix="classes")
def get_token_class(tokentype):
    """
    This takes a token type like "hotp" and returns a class
    like <class privacidea.lib.tokens.hotptoken.HotpTokenClass>
    :return: The tokenclass for the given type
    :rtype: tokenclass.TokenClass
    """
    if tokentype.lower() == "hmac":
        tokentype = "hotp"

    tokenclass = None
    tokenclasses = get_token_classes()
    for tclass in tokenclasses:
        if tclass.get_class_type().lower() == tokentype.lower():
            tokenclass = tclass
            break

    return tokenclass


# @cache.cached(key_prefix="types")
def get_token_types():
    """
    Return a simple list of the type names of the tokens.
    :return: list of tokentypes like 'hotp', 'totp'...
    :rtype: list
    """
    if "pi_token_types" not in this.config:
        (t_classes, t_types) = get_token_class_dict()
        this.config["pi_token_types"] = t_types
        this.config["pi_token_classes"] = t_classes

    return list(this.config["pi_token_types"].values())


# @cache.cached(key_prefix="prefix")
def get_token_prefix(tokentype=None, default=None):
    """
    Return the token prefix for a tokentype as it is defined in the
    tokenclass. If no tokentype is specified, we return a dictionary
    with the tokentypes as keys.
    :param tokentype: the type of the token like "hotp" or "totp"
    :type tokentype: basestring
    :param default: If the tokentype is not found, we return default
    :type default: basestring
    :return: the prefix of the tokentype or the dict with all prefixes
    :rtype: string or dict
    """
    prefix_dict = {}
    for tokenclass in get_token_classes():
        prefix_dict[tokenclass.get_class_type()] = tokenclass.get_class_prefix()

    if tokentype:
        ret = prefix_dict.get(tokentype, default)
    else:
        ret = prefix_dict
    return ret


# @cache.cached(key_prefix="classes")
def get_token_classes():
    """
    Returns a list of the available token classes like:
    [<class 'privacyidea.lib.tokens.totptoken.TotpTokenClass'>,
    <class 'privacyidea.lib.tokens.hotptoken.HotpTokenClass'>]

    :return: array of token classes
    :rtype: list
    """
    if "pi_token_classes" not in this.config:
        (t_classes, t_types) = get_token_class_dict()
        this.config["pi_token_classes"] = t_classes
        this.config["pi_token_types"] = t_types

    return list(this.config["pi_token_classes"].values())


def get_machine_resolver_class_dict():
    """
    get a dictionary of the machine resolver classes and a dictionary of the
    machines resolver types like this:

    ({'privacyidea.lib.machines.hosts.HostsMachineResolver':
      <class 'privacyidea.lib.machines.hosts.HostsMachineResolver'>},
     {'privacyidea.lib.machines.hosts.HostsMachineResolver':
      'hosts'})

    :return: tuple of two dicts
    """
    resolverclass_dict = {}
    resolvertype_dict = {}

    modules = get_machine_resolver_module_list()
    for module in modules:
        log.debug("module: {0!s}".format(module))
        for name in dir(module):
            obj = getattr(module, name)
            if inspect.isclass(obj) and \
                    (issubclass(obj, BaseMachineResolver)) and \
                    (obj != BaseMachineResolver):
                try:
                    class_name = "{0!s}.{1!s}".format(module.__name__, obj.__name__)
                    resolverclass_dict[class_name] = obj
                    resolvertype_dict[class_name] = obj.type

                except Exception as e:  # pragma: no cover
                    log.error("error constructing machine resolver "
                              "class_list: %r" % e)

    return resolverclass_dict, resolvertype_dict


def get_caconnector_class_dict():
    """
    get a dictionary of the CA connector classes and a dictionary of the
    machines resolver types like this:

    ({'privacyidea.lib.caconnectors.localca.LocalCAConnector':
      <class 'privacyidea.lib.caconnectors.localca.LocalCAConnector'>},
     {'privacyidea.lib.caconnectors.localca.LocalCAConnector':
      'local'})

    :return: tuple of two dicts
    """
    class_dict = {}
    type_dict = {}

    modules = get_caconnector_module_list()
    for module in modules:
        log.debug("module: {0!s}".format(module))
        for name in dir(module):
            obj = getattr(module, name)
            if inspect.isclass(obj) and \
                    (issubclass(obj, BaseCAConnector)) and \
                    (obj != BaseCAConnector):
                try:
                    class_name = "{0!s}.{1!s}".format(module.__name__, obj.__name__)
                    class_dict[class_name] = obj
                    type_dict[class_name] = obj.connector_type

                except Exception as e:  # pragma: no cover
                    log.error("error constructing CA connector "
                              "class_list: %r" % e)

    return class_dict, type_dict


# @cache.cached(key_prefix="resolver")
def get_resolver_class_dict():
    """
    get a dictionary of the resolver classes and a dictionary
    of the resolver types:

    ({'privacyidea.lib.resolvers.PasswdIdResolver.IdResolver':
      <class 'privacyidea.lib.resolvers.PasswdIdResolver.IdResolver'>,
      'privacyidea.lib.resolvers.PasswdIdResolver.UserIdResolver':
      <class 'privacyidea.lib.resolvers.UserIdResolver.UserIdResolver'>},

      {'privacyidea.lib.resolvers.PasswdIdResolver.IdResolver':
      'passwdresolver',
      'privacyidea.lib.resolvers.PasswdIdResolver.UserIdResolver':
      'UserIdResolver'})

    :return: tuple of two dicts.
    """
    resolverclass_dict = {}
    resolverprefix_dict = {}

    modules = get_resolver_module_list()
    for module in modules:
        log.debug("module: {0!s}".format(module))
        for name in dir(module):
            obj = getattr(module, name)
            # There are other classes like HMAC in the lib.tokens module,
            # which we do not want to load.
            if inspect.isclass(obj) and (issubclass(obj, UserIdResolver) or
                                         obj == UserIdResolver):
                # We must not process imported classes!
                # if obj.__module__ == module.__name__:
                try:
                    class_name = "{0!s}.{1!s}".format(module.__name__, obj.__name__)
                    resolverclass_dict[class_name] = obj

                    prefix = class_name.split('.')[1]
                    if hasattr(obj, 'getResolverClassType'):
                        prefix = obj.getResolverClassType()

                    resolverprefix_dict[class_name] = prefix

                except Exception as e:  # pragma: no cover
                    log.error("error constructing resolverclass_list: {0!r}".format(e))

    return resolverclass_dict, resolverprefix_dict


@log_with(log)
# @cache.cached(key_prefix="resolver")
def get_resolver_list():
    """
    get the list of the module names of the resolvers like
    "resolvers.PasswdIdResolver".

    :return: list of resolver names from the config file
    :rtype: set
    """
    module_list = set()

    module_list.add("privacyidea.lib.resolvers.PasswdIdResolver")
    module_list.add("privacyidea.lib.resolvers.LDAPIdResolver")
    module_list.add("privacyidea.lib.resolvers.SCIMIdResolver")
    module_list.add("privacyidea.lib.resolvers.SQLIdResolver")
    module_list.add("privacyidea.lib.resolvers.HTTPResolver")

    # Dynamic Resolver modules
    # TODO: Migration
    # config_modules = config.get("privacyideaResolverModules", '')
    config_modules = None
    log.debug("{0!s}".format(config_modules))
    if config_modules:
        # in the config *.ini files we have some line continuation slashes,
        # which will result in ugly module names, but as they are followed by
        # \n they could be separated as single entries by the following two
        # lines
        lines = config_modules.splitlines()
        coco = ",".join(lines)
        for module in coco.split(','):
            if module.strip() != '\\':
                module_list.add(module.strip())

    return module_list


@log_with(log)
# @cache.memoize(1)
def get_machine_resolver_class_list():
    """
    get the list of the class names of the machine resolvers like
    "machines.hosts.HostsMachineResolver".

    :return: list of machine resolver class names from the config file
    :rtype: list
    """
    class_list = []
    # TODO: We should read all classes inherited by BaseMachineResolver under
    # machines/
    # Otherwise we need to add a line here for each new MachineResolver.
    class_list.append("privacyidea.lib.machines.hosts.HostsMachineResolver")
    class_list.append("privacyidea.lib.machines.ldap.LdapMachineResolver")
    return class_list


@log_with(log)
# @cache.cached(key_prefix="token")
def get_token_list():
    """
    get the list of the tokens
    :return: list of token names from the config file
    """
    module_list = set()

    module_list.add("privacyidea.lib.tokens.daplugtoken")
    module_list.add("privacyidea.lib.tokens.hotptoken")
    module_list.add("privacyidea.lib.tokens.motptoken")
    module_list.add("privacyidea.lib.tokens.passwordtoken")
    module_list.add("privacyidea.lib.tokens.applicationspecificpasswordtoken")
    module_list.add("privacyidea.lib.tokens.remotetoken")
    module_list.add("privacyidea.lib.tokens.daypasswordtoken")
    module_list.add("privacyidea.lib.tokens.spasstoken")
    module_list.add("privacyidea.lib.tokens.sshkeytoken")
    module_list.add("privacyidea.lib.tokens.totptoken")
    module_list.add("privacyidea.lib.tokens.yubicotoken")
    module_list.add("privacyidea.lib.tokens.yubikeytoken")
    module_list.add("privacyidea.lib.tokens.radiustoken")
    module_list.add("privacyidea.lib.tokens.smstoken")
    module_list.add("privacyidea.lib.tokens.emailtoken")
    module_list.add("privacyidea.lib.tokens.registrationtoken")
    module_list.add("privacyidea.lib.tokens.certificatetoken")
    module_list.add("privacyidea.lib.tokens.foureyestoken")
    module_list.add("privacyidea.lib.tokens.tiqrtoken")
    module_list.add("privacyidea.lib.tokens.ocratoken")
    module_list.add("privacyidea.lib.tokens.u2ftoken")
    module_list.add("privacyidea.lib.tokens.papertoken")
    module_list.add("privacyidea.lib.tokens.questionnairetoken")
    module_list.add("privacyidea.lib.tokens.vascotoken")
    module_list.add("privacyidea.lib.tokens.tantoken")
    module_list.add("privacyidea.lib.tokens.pushtoken")
    module_list.add("privacyidea.lib.tokens.indexedsecrettoken")
    module_list.add("privacyidea.lib.tokens.webauthntoken")
    module_list.add("privacyidea.lib.tokens.passkeytoken")

    # Dynamic token modules
    dynamic_token_modules = get_app_config_value("PI_TOKEN_MODULES")
    if dynamic_token_modules:
        # In the pi.cfg you can specify a list or set of 3rd party token modules like
        #    PI_TOKEN_MODULES = [ "myproj.tokens.tok1", "myproj.tokens.tok2" ]
        module_list.update(to_list(dynamic_token_modules))

    return module_list


def get_email_validators():
    """
    Return a dict of available email validator modules and its functions

    The list be configured in "PI_EMAIL_VALIDATOR_MODULES" in pi.cfg.

    The return value looks like:
    { "privacyidea.lib.utils.emailvalidation": func:validate_email }

    :return: a dict
    """
    if "pi_email_validators" not in this.config:
        validator_dict = {}
        validator_list = to_list(get_app_config_value("PI_EMAIL_VALIDATOR_MODULES"))
        # Add our default validator
        validator_list.append("privacyidea.lib.utils.emailvalidation")
        # Remove empty entries
        validator_list = [x for x in validator_list if x]

        for mod_name in validator_list:
            if mod_name == '\\' or len(mod_name.strip()) == 0:
                continue
            try:
                log.debug("import module: {0!s}".format(mod_name))
                module = importlib.import_module(mod_name)
                validator_dict[mod_name] = module.validate_email
            except Exception as exx:  # pragma: no cover
                log.warning('unable to load validate module with function "validate_email": '
                            '{0!r} ({1!r})'.format(mod_name, exx))
        this.config["pi_email_validators"] = validator_dict

    return this.config["pi_email_validators"]


@log_with(log)
# @cache.cached(key_prefix="modules")
def get_token_module_list():
    """
    return the list of modules of the available token classes

    :return: list of token modules
    """
    # def load_resolver_modules
    module_list = get_token_list()
    log.debug("using the module list: {0!s}".format(module_list))

    modules = []
    for mod_name in module_list:
        if mod_name == '\\' or len(mod_name.strip()) == 0:
            continue

        try:
            log.debug("import module: {0!s}".format(mod_name))
            module = importlib.import_module(mod_name)
            modules.append(module)
        except Exception as exx:  # pragma: no cover
            module = None
            log.warning('unable to load token module : {0!r} ({1!r})'.format(mod_name, exx))

    return modules


# @cache.cached(key_prefix="modules")
def get_resolver_module_list():
    """
    return the list of modules of the available resolver classes
    like passwd, sql, ldap

    :return: list of resolver modules
    """

    # def load_resolver_modules
    module_list = get_resolver_list()
    log.debug("using the module list: {0!s}".format(module_list))

    modules = []
    for mod_name in module_list:
        if mod_name == '\\' or len(mod_name.strip()) == 0:
            continue

        try:
            log.debug("import module: {0!s}".format(mod_name))
            module = importlib.import_module(mod_name)

        except Exception as exx:  # pragma: no cover
            module = None
            log.warning('unable to load resolver module : {0!r} ({1!r})'.format(mod_name, exx))

        if module is not None:
            modules.append(module)

    return modules


# @cache.cached(key_prefix="module")
def get_caconnector_module_list():
    """
    return the list of modules of the available CA connector classes

    :return: list of CA connector modules
    """
    from privacyidea.lib.caconnectors.baseca import AvailableCAConnectors
    module_list = set(AvailableCAConnectors)

    modules = []
    for mod_name in module_list:
        mod_name = ".".join(mod_name.split(".")[:-1])
        class_name = mod_name.split(".")[-1:]
        try:
            log.debug("import module: {0!s}".format(mod_name))
            module = importlib.import_module(mod_name)

        except Exception as exx:  # pragma: no cover
            module = None
            log.warning('unable to load ca connector module : {0!r} ({1!r})'.format(mod_name, exx))

        if module is not None:
            modules.append(module)

    return modules


# @cache.cached(key_prefix="module")
def get_machine_resolver_module_list():
    """
    return the list of modules of the available machines resolver classes
    like base, hosts

    :return: list of resolver modules
    """

    # def load_resolver_modules
    class_list = get_machine_resolver_class_list()
    log.debug("using the class list: {0!s}".format(class_list))

    modules = []
    for class_name in class_list:
        try:
            module_name = ".".join(class_name.split(".")[:-1])
            log.debug("import module: {0!s}".format(module_name))
            module = importlib.import_module(module_name)

        except Exception as exx:  # pragma: no cover
            module = None
            log.warning('unable to load machine resolver module : {0!r} ({1!r})'.format(module_name, exx))

        if module is not None:
            modules.append(module)

    return modules


def set_privacyidea_config(key, value, typ="", desc=""):
    """
    Set a config value and writes it to the Config database table.
    Can be of type "password" or "public". "password" gets encrypted.
    """
    if not typ:
        # check if this is a token specific config and if it should be public
        try:
            token_type = key.split(".")[0]
            tclass = get_token_class(token_type)
            typ = tclass.get_setting_type(key)
        except Exception:
            log.debug("This seems to be no token specific setting")

    if typ == "password":
        # store value in encrypted way
        value = encryptPassword(value)
    # We need to check, if the value already exist
    c1 = Config.query.filter_by(Key=key).first()
    if c1:
        # The value already exist, we need to update
        c1.Value = value
        if typ:
            c1.Type = typ
        if desc:
            c1.Description = desc
        save_config_timestamp()
        db.session.commit()
        ret = "update"
    else:
        # new_entry = Config(key, value, typ, desc)
        # ``save`` will call ``save_config_timestamp`` for us
        Config(key, value, typ, desc).save()
        # db.session.add(new_entry)
        ret = "insert"

    # save_config_timestamp()
    # db.session.commit()
    return ret


def delete_privacyidea_config(key):
    """
    Delete a config entry
    """
    ret = 0
    # We need to check, if the value already exist
    if Config.query.filter_by(Key=key).first().delete():
        ret = True
    # if q:
    #    db.session.delete(q)
    #    db.session.commit()
    #    ret = True
    #    save_config_timestamp()
    return ret


# @cache.cached(key_prefix="pin")
def get_inc_fail_count_on_false_pin():
    """
    Return if the Failcounter should be increased if only tokens
    with a false PIN were identified.
    :return: True of False
    :rtype: bool
    """
    r = get_from_config(key="IncFailCountOnFalsePin",
                        default=False, return_bool=True)
    return r


# @cache.cached(key_prefix="pin")
def get_prepend_pin():
    """
    Get the status of the "PrependPin" Config

    :return: True or False
    :rtype: bool
    """
    r = get_from_config(key="PrependPin", default=False, return_bool=True)
    return r


def set_prepend_pin(prepend=True):
    """
    Set the status of the "PrependPin" Config
    :param prepend: If the PIN should be prepended or not
    :return: None
    """
    set_privacyidea_config("PrependPin", prepend)


def return_saml_attributes():
    r = get_from_config(key=SYSCONF.RETURNSAML, default=False,
                        return_bool=True)
    return r


def return_saml_attributes_on_fail():
    r = get_from_config(key=SYSCONF.RETURNSAMLONFAIL, default=False,
                        return_bool=True)
    return r


def get_privacyidea_node(default='localnode') -> str:
    """
    This returns the node name of the privacyIDEA node as found in the pi.cfg
    file in PI_NODE.
    If it does not exist, the PI_AUDIT_SERVERNAME is used.

    :return: the distinct node name
    :rtype: str
    """
    node_name = get_app_config_value("PI_NODE", get_app_config_value("PI_AUDIT_SERVERNAME", default))
    return node_name


def get_privacyidea_nodes() -> list:
    """
    This returns the list of the nodes known to privacyIDEA.
    It includes the own local node name and returns the name and the
    corresponding UUID

    :return: list of nodes as dicts with name and uuid
    :rtype: list
    """
    nodes_list = []
    nodes = db.session.query(NodeName).all()

    for node in nodes:
        nodes_list.append({
            "uuid": node.id,
            "name": node.name})

    return nodes_list


def get_privacyidea_node_names() -> list:
    """
    This returns the list of the node names known to privacyIDEA.
    It includes the own local node name.

    :return: list of node names as strings
    :rtype: list
    """
    return [node.get("name") for node in get_privacyidea_nodes()]


def check_node_uuid_exists(node_uuid) -> bool:
    """
    Check if a node with the given UUID exists in the database.

    :param node_uuid: the UUID of the node
    :return: True if the node exists in the database, False otherwise
    """
    return db.session.query(NodeName).filter_by(id=node_uuid).first() is not None


@register_export()
def export_config(name=None):
    """Export the global configuration"""
    c = copy.copy(get_config_object().config)
    if name:
        c = {name: c[name]} if name in c.keys() else {}
    return c


@register_import()
def import_config(data, name=None):
    """Import given server configuration"""
    log.debug('Import server config: {0!s}'.format(data))
    res = {}
    data.pop('__timestamp__', None)
    for key, values in data.items():
        if name and name != key:
            continue
        r = set_privacyidea_config(key, values['Value'],
                                   desc=values['Description'] if 'Description' in values else None,
                                   typ=values['Type'] if 'Type' in values else None)
        res[key] = r
    log.info('Added configuration: {0!s}'.format(
        ', '.join([k for k, v in res.items() if v == 'insert'])))
    log.info('Updated configuration: {0!s}'.format(
        ', '.join([k for k, v in res.items() if v == 'update'])))


def get_multichallenge_enrollable_tokentypes():
    enrollable_tokentypes = []
    # If the token is enrollable via multichallenge
    for tclass in get_token_classes():
        if tclass.is_multichallenge_enrollable():
            enrollable_tokentypes.append(tclass.get_class_type())
    return enrollable_tokentypes



"""
2017-12-27 Cornelius Kölbel <cornelius.koelbel@netknights.it>
           Mocking the SMPP client


Original responses.py is:
Copyright 2013 Dropbox, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import namedtuple
from mock import Mock

from collections.abc import Sequence, Sized

from smpplib.exceptions import ConnectionError

from .smtpmock import get_wrapped

Call = namedtuple('Call', ['request', 'response'])

_wrapper_template = """\
def wrapper%(signature)s:
    with smppmock:
        return func%(funcargs)s
"""


class CallList(Sequence, Sized):
    def __init__(self):
        self._calls = []

    def __iter__(self):
        return iter(self._calls)

    def __len__(self):
        return len(self._calls)

    def __getitem__(self, idx):
        return self._calls[idx]

    def setdata(self, request, response):
        self._calls.append(Call(request, response))

    def reset(self):
        self._calls = []


class SmppMock(object):

    def __init__(self):
        self._calls = CallList()
        self.connect_successful = True
        self.systemid = None
        self.password = None
        self.reset()

    def reset(self):
        self._calls.reset()

    def setdata(self, connection_success=True, systemid=None, password=None):
        self.connect_successful = connection_success
        self.systemid = systemid
        self.password = password

    @property
    def calls(self):
        return self._calls

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop()
        self.reset()

    def activate(self, func):
        evaldict = {'smppmock': self, 'func': func}
        return get_wrapped(func, _wrapper_template, evaldict)

    def _on_connect(self, SMPP):
        if self.connect_successful:
            return None
        else:
            raise ConnectionError()

    def _on_disconnect(self, SMPP):
        self.host = None
        self.port = None

    def _on_init(self, SMPP, host, port):
        self.host = host
        self.port = port
        return None

    def _on_transmitter(self, SMPP, systemid, password):
        if systemid != self.systemid or password != self.password:
            raise Exception("Wrong credentials")
        BindTransmitterRespMock = Mock()
        BindTransmitterRespMock.get_status_desc.return_value = 'No Error'
        return BindTransmitterRespMock

    def _on_send_message(self, SMPP, source_addr_ton=None, source_addr_npi=None,
                         source_addr=None, dest_addr_ton=None,
                         dest_addr_npi=None, destination_addr=None,
                         short_message=None, data_coding=None, esm_class=None):
        SubmitSMMock = Mock()
        SubmitSMMock.get_status_desc.return_value = 'No Error'
        return SubmitSMMock

    def start(self):
        import mock

        def unbound_on_init(SMPP, host, port):
            return self._on_init(SMPP, host, port)

        self._patcher1 = mock.patch('smpplib.client.Client.__init__',
                                   unbound_on_init)
        self._patcher1.start()

        def unbound_on_connect(SMPP):
            return self._on_connect(SMPP)

        self._patcher2 = mock.patch('smpplib.client.Client.connect',
                                    unbound_on_connect)
        self._patcher2.start()

        def unbound_on_transmitter(SMPP, system_id, password):
            return self._on_transmitter(SMPP, system_id, password)

        self._patcher3 = mock.patch('smpplib.client.Client.bind_transmitter',
                                    unbound_on_transmitter)
        self._patcher3.start()

        def unbound_on_send_message(SMPP, *a, **kwargs):
            return self._on_send_message(SMPP, *a, **kwargs)

        self._patcher4 = mock.patch('smpplib.client.Client.send_message',
                                    unbound_on_send_message)
        self._patcher4.start()

        def unbound_on_disconnect(SMPP):
            return self._on_disconnect(SMPP)

        self._patcher5 = mock.patch('smpplib.client.Client.disconnect',
                                    unbound_on_disconnect)
        self._patcher5.start()

    def stop(self):
        self._patcher1.stop()
        self._patcher2.stop()
        self._patcher3.stop()
        self._patcher4.stop()
        self._patcher5.stop()


# expose default mock namespace
mock = _default_mock = SmppMock()
__all__ = []
for __attr in (a for a in dir(_default_mock) if not a.startswith('_')):
    __all__.append(__attr)
    globals()[__attr] = getattr(_default_mock, __attr)



"""v3.10: Add index to expiration in challenge table

Revision ID: 2100d1fad908
Revises: 250931d82e51
Create Date: 2024-05-06 13:36:45.910206

"""

# revision identifiers, used by Alembic.
revision = '2100d1fad908'
down_revision = '250931d82e51'

from alembic import op
from sqlalchemy.exc import OperationalError, ProgrammingError


def upgrade():
    try:
        op.create_index(op.f('ix_challenge_expiration'), 'challenge', ['expiration'], unique=False)
    except (OperationalError, ProgrammingError) as exx:
        if "Index already exists" in str(exx.orig).lower():
            print("Ok, Index 'expiration' in table 'challenge' already exists.")
        else:
            print(exx)
    except Exception as exx:
        print("Could not create index 'expiration' on table 'challenge'")
        print(exx)


def downgrade():
    op.drop_index(op.f('ix_challenge_expiration'), table_name='challenge')



#  2019-02-04 Friedrich Weber <friedrich.weber@netknights.it>
#             Add a job queue
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import logging

from flask import current_app

from privacyidea.lib.error import ServerError
from privacyidea.lib.framework import get_app_local_store
from privacyidea.lib.utils import get_module_class

log = logging.getLogger(__name__)

JOB_QUEUE_CLASS = "PI_JOB_QUEUE_CLASS"
JOB_QUEUE_OPTION_PREFIX = "PI_JOB_QUEUE_"


class JobCollector(object):
    """
    For most third-party job queue modules, the jobs are discovered by tracking all
    functions decorated with a ``@job`` decorator. However, in order
    to invoke the decorator, one usually needs to provide the queue
    configuration (e.g. the redis server) already.
    In privacyIDEA, we cannot do that, because the app config is not known
    yet -- it will be known when ``create_app`` is called!
    Thus, we cannot directly use the @job decorator, but need a job collector
    that collects jobs in privacyIDEA code and registers them with the job
    queue module when ``create_app`` has been called.
    """
    def __init__(self):
        self._jobs = {}

    @property
    def jobs(self):
        return self._jobs

    def register_job(self, name, func, args, kwargs):
        """
        Register a job with the collector.

        :param name: unique name of the job
        :param func: function of the job
        :param args: arguments passed to the job queue's ``register_job`` method
        :param kwargs: keyword arguments passed to the job queue's ``register_job`` method
        """
        if name in self._jobs:
            raise RuntimeError("Duplicate jobs: {!r}".format(name))
        self._jobs[name] = (func, args, kwargs)

    def register_app(self, app):
        """
        Create an instance of a ``BaseQueue`` subclass according to the app config's
        ``PI_JOB_QUEUE_CLASS`` option and store it in the ``job_queue`` config.
        Register all collected jobs with this application.
        This instance is shared between threads!
        This function should only be called once per process.

        :param app: privacyIDEA app
        """
        with app.app_context():
            store = get_app_local_store()
        if "job_queue" in store:
            raise RuntimeError("App already has a job queue: {!r}".format(store["job_queue"]))
        try:
            package_name, class_name = app.config[JOB_QUEUE_CLASS].rsplit(".", 1)
            queue_class = get_module_class(package_name, class_name)
        except (ImportError, ValueError) as exx:
            log.warning("Could not import job queue class {!r}: {!r}".format(app.config[JOB_QUEUE_CLASS], exx))
            return
        # Extract configuration from app config: All options starting with PI_JOB_QUEUE_
        options = {}
        for k, v in app.config.items():
            if k.startswith(JOB_QUEUE_OPTION_PREFIX) and k != JOB_QUEUE_CLASS:
                options[k[len(JOB_QUEUE_OPTION_PREFIX):].lower()] = v
        job_queue = queue_class(options)
        log.info("Created a new job queue: {!r}".format(job_queue))
        store["job_queue"] = job_queue
        for name, (func, args, kwargs) in self._jobs.items():
            job_queue.register_job(name, func, *args, **kwargs)


#: A singleton is fine here, because it is only used at
#: import time and once when a new app is created.
#: Afterwards, the object is unused.
JOB_COLLECTOR = JobCollector()


def job(name, *args, **kwargs):
    """
    Decorator to mark a job to be collected by the job collector.
    All arguments are passed to ``register_job``.
    """
    def decorator(f):
        JOB_COLLECTOR.register_job(name, f, args, kwargs)
        return f
    return decorator


def register_app(app):
    """
    Register the app ``app`` with the global job collector, if a PI_JOB_QUEUE_CLASS is non-empty.
    Do nothing otherwise.
    """
    if app.config.get(JOB_QUEUE_CLASS, ""):
        JOB_COLLECTOR.register_app(app)


def has_job_queue():
    """
    Return a boolean describing whether the current app has an app queue configured.
    """
    return "job_queue" in get_app_local_store()


def get_job_queue():
    """
    Get the job queue registered with the current app. If no job queue is configured,
    raise a ServerError.
    """
    store = get_app_local_store()
    if "job_queue" in store:
        return store["job_queue"]
    else:
        raise ServerError("privacyIDEA has no job queue configured!")


def wrap_job(name, result):
    """
    Wrap a job and return a function that can be used like the original function.
    The returned function will always return ``result``.
    This assumes that a queue is configured! Otherwise, calling the
    resulting function will fail with a ServerError.

    :return: a function
    """
    def caller(*args, **kwargs):
        get_job_queue().enqueue(name, args, kwargs)
        return result
    return caller



"""
This testfile tests the basic app functionality of the privacyIDEA app
"""
import os
import unittest
import flask
import inspect
import logging
import mock
from testfixtures import Comparison, compare
from privacyidea.app import create_app
from privacyidea.config import config, TestingConfig

dirname = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))


class AppTestCase(unittest.TestCase):
    def setUp(self):
        self.logger = logging.getLogger()
        self.orig_handlers = self.logger.handlers
        self.logger.handlers = []
        self.level = self.logger.level

    def tearDown(self):
        self.logger.handlers = self.orig_handlers
        self.logger.level = self.level

    def test_01_create_default_app(self):
        # This will create the app with the 'development' configuration
        app = create_app()
        self.assertIsInstance(app, flask.app.Flask, app)
#        self.assertEqual(app.env, 'production', app)
        self.assertTrue(app.debug, app)
        self.assertFalse(app.testing, app)
        self.assertEqual(app.import_name, 'privacyidea.app', app)
        self.assertEqual(app.name, 'privacyidea.app', app)
#        self.assertTrue(app.response_class == PiResponseClass, app)
        blueprints = ['validate_blueprint', 'token_blueprint', 'system_blueprint',
                      'resolver_blueprint', 'realm_blueprint', 'defaultrealm_blueprint',
                      'policy_blueprint', 'login_blueprint', 'jwtauth', 'user_blueprint',
                      'audit_blueprint', 'machineresolver_blueprint', 'machine_blueprint',
                      'application_blueprint', 'caconnector_blueprint', 'cert_blueprint',
                      'ttype_blueprint', 'register_blueprint', 'smtpserver_blueprint',
                      'recover_blueprint', 'radiusserver_blueprint', 'periodictask_blueprint',
                      'privacyideaserver_blueprint', 'eventhandling_blueprint',
                      'smsgateway_blueprint', 'client_blueprint', 'subscriptions_blueprint',
                      'monitoring_blueprint']
        self.assertTrue(all(k in app.before_request_funcs for k in blueprints), app)
        self.assertTrue(all(k in app.blueprints for k in blueprints), app)
        extensions = ['sqlalchemy', 'migrate', 'babel']
        self.assertTrue(all(k in extensions for k in app.extensions), app)
        self.assertEqual(app.secret_key, 't0p s3cr3t', app)
        # TODO: check url_map and view_functions
        # check that the configuration was loaded successfully
        # the default configuration is 'development'
        dc = config['development']()
        members = inspect.getmembers(dc, lambda a: not (inspect.isroutine(a)))
        conf = [m for m in members if not (m[0].startswith('__') and m[0].endswith('__'))]
        self.assertTrue(all(app.config[k] == v for k, v in conf), app)
        # check the correct initialization of the logging
        logger = logging.getLogger('privacyidea')
        self.assertEqual(logger.level, logging.DEBUG, logger)
        compare([
            Comparison('logging.handlers.RotatingFileHandler',
                       baseFilename=os.path.join(dirname, 'privacyidea.log'),
                       formatter=Comparison('privacyidea.lib.log.SecureFormatter',
                                            _fmt="[%(asctime)s][%(process)d]"
                                                 "[%(thread)d][%(levelname)s]"
                                                 "[%(name)s:%(lineno)d] "
                                                 "%(message)s",
                                            partial=True),
                       level=logging.DEBUG,
                       partial=True)
        ], logger.handlers)

    def test_02_create_production_app(self):
        app = create_app(config_name='production')
        dc = config['production']()
        members = inspect.getmembers(dc, lambda a: not (inspect.isroutine(a)))
        conf = [m for m in members if not (m[0].startswith('__') and m[0].endswith('__'))]
        self.assertTrue(all(app.config[k] == v for k, v in conf), app)

    def test_03_logging_config_file(self):
        class Config(TestingConfig):
            PI_LOGCONFIG = "tests/testdata/logging.cfg"
        with mock.patch.dict("privacyidea.config.config", {"testing": Config}):
            create_app(config_name='testing')
            # check the correct initialization of the logging from config file
            logger = logging.getLogger('privacyidea')
            self.assertEqual(logger.level, logging.DEBUG, logger)
            compare([
                Comparison('logging.handlers.RotatingFileHandler',
                           baseFilename=os.path.join(dirname, 'privacyidea.log'),
                           formatter=Comparison('privacyidea.lib.log.SecureFormatter',
                                                _fmt="[%(asctime)s][%(process)d]"
                                                     "[%(thread)d][%(levelname)s]"
                                                     "[%(name)s:%(lineno)d] "
                                                     "%(message)s",
                                                partial=True),
                           level=logging.DEBUG,
                           partial=True)
            ], logger.handlers)
            logger = logging.getLogger('privacyidea.lib.auditmodules.loggeraudit')
            self.assertEqual(logger.level, logging.INFO, logger)
            compare([
                Comparison('logging.handlers.RotatingFileHandler',
                           baseFilename=os.path.join(dirname, 'audit.log'),
                           formatter=Comparison('privacyidea.lib.log.SecureFormatter',
                                                _fmt="[%(asctime)s][%(process)d]"
                                                     "[%(thread)d][%(levelname)s]"
                                                     "[%(name)s:%(lineno)d] "
                                                     "%(message)s",
                                                partial=True),
                           level=logging.INFO,
                           partial=True)
            ], logger.handlers)

    def test_04_logging_config_yaml(self):
        class Config(TestingConfig):
            PI_LOGCONFIG = "tests/testdata/logging.yml"
        with mock.patch.dict("privacyidea.config.config", {"testing": Config}):
            create_app(config_name='testing')
            # check the correct initialization of the logging from config file
            logger = logging.getLogger('privacyidea')
            self.assertEqual(logger.level, logging.INFO, logger)
            compare([
                Comparison('logging.handlers.RotatingFileHandler',
                           baseFilename=os.path.join(dirname, 'privacyidea.log'),
                           formatter=Comparison('privacyidea.lib.log.SecureFormatter',
                                                _fmt="[%(asctime)s][%(process)d]"
                                                     "[%(thread)d][%(levelname)s]"
                                                     "[%(name)s:%(lineno)d] "
                                                     "%(message)s",
                                                partial=True),
                           backupCount=5,
                           level=logging.DEBUG,
                           partial=True)
            ], logger.handlers)
            logger = logging.getLogger('audit')
            self.assertEqual(logger.level, logging.INFO, logger)
            compare([
                Comparison('logging.handlers.RotatingFileHandler',
                           backupCount=14,
                           baseFilename=os.path.join(dirname, 'audit.log'),
                           level=logging.INFO,
                           formatter=None,
                           partial=True)
            ], logger.handlers)

    def test_05_logging_config_broken_yaml(self):
        class Config(TestingConfig):
            PI_LOGCONFIG = "tests/testdata/logging_broken.yaml"
        with mock.patch.dict("privacyidea.config.config", {"testing": Config}):
            create_app(config_name='testing')
            # check the correct initialization of the logging with the default
            # values since the yaml file is broken
            logger = logging.getLogger('privacyidea')
            self.assertEqual(logger.level, logging.INFO, logger)
            compare([
                Comparison('logging.handlers.RotatingFileHandler',
                           baseFilename=os.path.join(dirname, 'privacyidea.log'),
                           formatter=Comparison('privacyidea.lib.log.SecureFormatter',
                                                _fmt="[%(asctime)s][%(process)d]"
                                                     "[%(thread)d][%(levelname)s]"
                                                     "[%(name)s:%(lineno)d] "
                                                     "%(message)s",
                                                partial=True),
                           level=logging.INFO,
                           partial=True)
            ], logger.handlers)
